#include<stdio.h>
#include<malloc.h>
#include<math.h>

#include "cuda_utils.h"

const int iterations=20;

// iterate function for computation of reference //

static inline float iterate(float value) {
  for(int j=0;j<iterations;j++) {
    value=4.f*value*(1.f-value);
  }
  return value;
}

// CUDA Kernel //

void __global__ compute(float* input,float* output,int size) {
  for(int i=threadIdx.x+blockIdx.x*blockDim.x;i<size;i+=blockDim.x*gridDim.x) {
    float value=input[i];
    for(int j=0;j<iterations;j++) {
      value=4.f*value*(1.f-value);
    }
    output[i]=value;
  }
}


// CPU code //
int main() {

  srand48(time(NULL));

  // define vector size //
  const int vector_size = 10000000;

  // number of streams or blocks //
  const int stream_count=8;

  // allocate memory on cpu //
  float* vector_input_cpu;
  float* vector_output_cpu;
  cu_verify(cudaMallocHost((void**)&vector_input_cpu,
                           vector_size*sizeof(float)));
  cu_verify(cudaMallocHost((void**)&vector_output_cpu,
                           vector_size*sizeof(float)));

  // initialize cpu input vector //
  for(int i=0;i<vector_size;i++) {
    vector_input_cpu[i]=drand48();
  }
  // initialize cpu output vector to 0 //
  for(int i=0;i<vector_size;i++) {
    vector_output_cpu[i]=0;
  }

  // allocate memory on gpu //
  float* vector_input_gpu;
  float* vector_output_gpu;
  cu_verify(cudaMalloc((void**)&vector_input_gpu,vector_size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&vector_output_gpu,vector_size*sizeof(float)));

  // init streams //
  // TODO: create CUDA streams //

  Timer timer;
  initTimer(&timer);

  // TODO: add loop over all streams //
  // TODO: change cudaMemcpy to cudaMemcpyAsync //
  // TODO: use correct start addresses and size for current vector segment
  // copy input vector to gpu //
  cu_verify(cudaMemcpy(vector_input_gpu,vector_input_cpu,
                       vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));

  // TODO: add loop over all streams //
  // TODO: attach kernel to stream //
  // TODO: use correct start addresses and size for current vector segment
  // cuda kernel call //
  cu_verify_void((compute<<<30,32>>>(vector_input_gpu,
                                     vector_output_gpu,
                                     vector_size)));

  // TODO: add loop over all streams //
  // TODO: change cudaMemcpy to cudaMemcpyAsync //
  // TODO: use correct start addresses and size for current vector segment
  // copy output vector from gpu to cpu //
  cu_verify(cudaMemcpy(vector_output_cpu,vector_output_gpu,
                       vector_size*sizeof(float),
                       cudaMemcpyDeviceToHost));
  
  // synchronize work //
  cudaThreadSynchronize();

  double duration = getTimer(&timer);
  printf("duration=%le seconds\n",
         duration);

  // free streams //
  // TODO: destroy streams //

  // verify results //
  bool success=true;
  for(int i=0;i<vector_size;i++) {
    if (fabsf(vector_output_cpu[i]-iterate(vector_input_cpu[i]))>1e-5) {
      printf("computation result is wrong: index=%i expected result=%g "
             "cuda result=%g\n",
             i,iterate(vector_input_cpu[i]),vector_output_cpu[i]);
      success=false;
      break;
    }
  }
  if (success) {
    printf("computation result is correct.\n");
  }

  // free gpu memory //
  cu_verify(cudaFree(vector_input_gpu));
  cu_verify(cudaFree(vector_output_gpu));

  // free cpu memory //
  cu_verify(cudaFreeHost(vector_input_cpu));
  cu_verify(cudaFreeHost(vector_output_cpu));

  // exit //
  return 0;
}
