module device_code

contains

! CUDA Kernel !

attributes(global) subroutine VectorSum(a,b,c,size,elements_per_thread)
  implicit none
  real, dimension(0:size-1) :: a,b,c
  integer, value :: size,elements_per_thread
  integer :: i,my_start,my_end
  
  my_start=((threadIdx%x-1)+(blockIdx%x-1)*blockDim%x)*elements_per_thread;
  my_end=((threadIdx%x-1)+(blockIdx%x-1)*blockDim%x+1)*elements_per_thread-1;
  if (my_end>size) then
    my_end=size
  end if
  do i=my_start,my_end
    c(i)=a(i)+b(i)
  end do
end subroutine VectorSum

end module device_code

program exercise04

use cudafor
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000000
integer, parameter :: block_dim = 64
integer, parameter :: grid_dim = 128
real, dimension(0:vector_size-1) :: vector_a_cpu
real, dimension(0:vector_size-1) :: vector_b_cpu
real, dimension(0:vector_size-1) :: vector_c_cpu
real, dimension(0:vector_size-1), device :: vector_a_gpu
real, dimension(0:vector_size-1), device :: vector_b_gpu
real, dimension(0:vector_size-1), device :: vector_c_gpu
real(kind=8) :: timer1,timer2,duration
integer :: elements_per_thread
integer :: i
logical :: success
integer :: error

! CPU code !

call random_seed()

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  call random_number(vector_a_cpu(i))
  call random_number(vector_b_cpu(i))
end do
! initialize cpu output vector to 0 !
do i=0,vector_size-1
   vector_c_cpu(i)=0;
end do

! copy input vectors a,b to gpu !
vector_a_gpu=vector_a_cpu
vector_b_gpu=vector_b_cpu
! initialize gpu output vector to 0 !
vector_c_gpu=vector_c_cpu

elements_per_thread = (vector_size+grid_dim*block_dim-1)/(grid_dim*block_dim)

call cpu_time(timer1)
! cuda kernel call !
call VectorSum<<<grid_dim,block_dim>>>(vector_a_gpu,vector_b_gpu,vector_c_gpu,vector_size,elements_per_thread)
error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! copy output vector from gpu to cpu !
vector_c_cpu=vector_c_gpu

! verify results !
success=.true.
do i=0,vector_size-1
  if (abs(vector_a_cpu(i)+vector_b_cpu(i)-vector_c_cpu(i))>1e-5) then
    print *,'computation result is wrong: index=',i, &
        ' expected result=',vector_a_cpu(i)*vector_b_cpu(i), &      
        'cuda result=', vector_c_cpu(i)
    success=.false.
    exit
  end if
end do
if (success) then
  print *,'computation result is correct.'
end if

print *,'duration=',duration,' seconds, performance = ',vector_size/duration,' FLOPS'
print *,'memory bandwidth = ',vector_size*12./duration,' bytes/s'

end program exercise04
